#ifdef _DEBUG
#include <ntddk.h>
#include "memtrack.h"

#define MAGIC	'TMEM'
#define INT_3	__asm int 3

struct prefix {
	ULONG magic;
	struct prefix *next, *prev;
	ULONG size;
	const char *file;
	ULONG line;
	char data[];
};

struct postfix {
	ULONG size;
	ULONG magic;
};

static KSPIN_LOCK guard;
static struct prefix *first, *last;
static ULONG count;

void memtrack_init()
{
	KeInitializeSpinLock(&guard);
}

void memtrack_free()
{
	KIRQL irql;
	ULONG total = 0;

	KeAcquireSpinLock(&guard, &irql);

	if (first) {
		struct prefix *p;
		for (p = first; p; p = p->next) {
			struct postfix *pd;

			if (p->magic != MAGIC) {
				KdPrint(("memtrack: memtrack_free: invalid pre-magic!\n"));
				INT_3;
				return;
			}

			pd = (struct postfix *)(p->data + p->size);

			if (pd->magic != MAGIC) {
				KdPrint(("memtrack: memtrack_free: invalid post-magic!\n"));
				INT_3;
				return;
			}

			if (p->size != pd->size) {
				KdPrint(("memtrack: memtracl_free: invalid post-size!\n"));
				INT_3;
				return;
			}

			KdPrint(("memtrack: memory leak detected! %d:%u (%u bytes)\n",
				p->file, p->line, p->size));

			total += p->size;
		}
	}

	KeReleaseSpinLock(&guard, irql);
	KdPrint(("memtrack: Total memory leakage: %u bytes (%u blocks)\n", total, count));
}

void *mt_malloc(ULONG size, const char *file, ULONG line)
{
	KIRQL irql;
	struct prefix *data = (struct prefix *)ExAllocatePool(NonPagedPool,
		sizeof(struct prefix) + size + sizeof(struct postfix));
	struct postfix *pd;

	if (!data) return NULL;

	data->magic = MAGIC;
	data->next = NULL;
	data->size = size;
	data->file = file;
	data->line = line;

	pd = (struct postfix *)(data->data + data->size);

	pd->size = size;
	pd->magic = MAGIC;

	KeAcquireSpinLock(&guard, &irql);
	
	if (last) {
		last->next = data;
		data->prev = last;
		last = data;
	}
	else {
		data->prev = NULL;
		first = last = data;
	}
	count++;

	KeReleaseSpinLock(&guard, irql);
	return data->data;
}

void free(void *ptr)
{
	KIRQL irql;
	struct prefix *data = (struct prefix *)((char *)ptr - sizeof(struct prefix));
	struct postfix *pd;
	
	if (data->magic != MAGIC) {
		KdPrint(("memtrack: free(0x%x): invalid pre-magic!\n", ptr));
		INT_3;
		return;
	}

	if (data->size == (ULONG)-1) {
		KdPrint(("memtrack: free(0x%x): double free of %s:%u!\n", ptr,
			data->file, data->line));
		INT_3;
		return;
	}

	pd = (struct postfix *)(data->data + data->size);

	if (pd->magic != MAGIC) {
		KdPrint(("memtrack: free(0x%x): invalid post-magic!\n", ptr));
		INT_3;
		return;
	}

	if (data->size != pd->size) {
		KdPrint(("memtrack: free(0x%x): invalid post-size!\n", ptr));
		INT_3;
		return;
	}

	KeAcquireSpinLock(&guard, &irql);

	if (data->next) data->next->prev = data->prev; else last = data->prev;
	if (data->prev) data->prev->next = data->next; else first = data->next;

	memset(data->data, 0xcc, data->size);

	data->size = (ULONG)-1;
	pd->size = (ULONG)-1;

	count--;
	KeReleaseSpinLock(&guard, irql);

	ExFreePool(data);
}

#endif
