// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
#include <ntddk.h>
#include <tdikrnl.h>
#include "memtrack.h"
#include "obj_tbl.h"

#define HASH_SIZE	0x1000
#define CALC_HASH(fileobj)  (((ULONG)(fileobj) >> 5) % HASH_SIZE)

static ot_entry_t **g_ot_hash;
KSPIN_LOCK g_ot_hash_guard;

typedef struct ctx_entry {
	struct ctx_entry *next;
	PFILE_OBJECT addrobj;
	CONNECTION_CONTEXT conn_ctx;
	PFILE_OBJECT connobj;
} ctx_entry_t;

static ctx_entry_t **g_cte_hash;
KSPIN_LOCK g_cte_hash_guard;

static NTSTATUS do_set_old_event_handler(ot_entry_t *ote, int event_type);

//----------------------------------------------------------------------------

NTSTATUS ot_init(void)
{
	g_ot_hash = (ot_entry_t **)malloc(sizeof(*g_ot_hash) * HASH_SIZE);
	if (!g_ot_hash) {
		KdPrint(("[tdi_flt] ot_init: malloc\n"));
		return STATUS_NO_MEMORY;
	}
	memset(g_ot_hash, 0, sizeof(*g_ot_hash) * HASH_SIZE);

	KeInitializeSpinLock(&g_ot_hash_guard);

	g_cte_hash = (ctx_entry_t **)malloc(sizeof(*g_cte_hash) * HASH_SIZE);
	if (!g_cte_hash) {
		KdPrint(("[tdi_flt] ot_init: malloc\n"));
		free(g_ot_hash);
		return STATUS_NO_MEMORY;
	}
	memset(g_cte_hash, 0, sizeof(*g_cte_hash) * HASH_SIZE);

	KeInitializeSpinLock(&g_cte_hash_guard);

	return STATUS_SUCCESS;
}

void ot_free(void)
{
	KIRQL irql;
	int i;

	if (g_ot_hash) {

		KeAcquireSpinLock(&g_ot_hash_guard, &irql);
		
		for (i = 0; i < HASH_SIZE; i++) {
			ot_entry_t *ote = g_ot_hash[i];
			while (ote) {
				ot_entry_t *ote2 = ote->next;
				int j;

				KdPrint(("[tdi_flt] ot_free: Warning! fileobj 0x%x type %d exists!\n",
					ote->fileobj, ote->type));
				
				if (ote->signature != 'OTE ')
					KdPrint(("[tdi_flt] ot_free: Warning! fileobj 0x%x invalid signature 0x%x!\n",
						ote->fileobj, ote->signature));

				for (j = 0; j < MAX_EVENT; j++)
					if (ote->ctx[j].old_handler) do_set_old_event_handler(ote, j);

				free(ote);
				ote = ote2;
			}
		}
		free(g_ot_hash); g_ot_hash = NULL;

		KeReleaseSpinLock(&g_ot_hash_guard, irql);
	}

	if (g_cte_hash) {
		KeAcquireSpinLock(&g_cte_hash_guard, &irql);
		
		for (i = 0; i < HASH_SIZE; i++) {
			ctx_entry_t *cte = g_cte_hash[i];
			while (cte) {
				ctx_entry_t *cte2 = cte->next;
				free(cte);
				cte = cte2;
			}
		}
		free(g_cte_hash); g_cte_hash = NULL;

		KeReleaseSpinLock(&g_cte_hash_guard, irql);
	}
}

//----------------------------------------------------------------------------

NTSTATUS ot_add_fileobj(PDEVICE_OBJECT devobj, PFILE_OBJECT fileobj, int fileobj_type,
						CONNECTION_CONTEXT conn_ctx)
{
	ULONG hash = CALC_HASH(fileobj);
	KIRQL irql;
	ot_entry_t *ote, *old_next;
	NTSTATUS status;
	int i;

	if (!fileobj) return STATUS_INVALID_PARAMETER_2;

	KeAcquireSpinLock(&g_ot_hash_guard, &irql);
	
	for (ote = g_ot_hash[hash]; ote; ote = ote->next)
		if (ote->fileobj == fileobj) break;

	if (!ote) {
		ote = (ot_entry_t *)malloc(sizeof(*ote));
		if (!ote) {
			KdPrint(("[tdi_flt] ot_add_fileobj: malloc\n"));
			status = STATUS_NO_MEMORY;
			goto done;
		}
		memset(ote, 0, sizeof(*ote));

		datapipe_init(&ote->in_oob_pipe);
		datapipe_init(&ote->in_pipe);
	
		ote->next = g_ot_hash[hash];
		g_ot_hash[hash] = ote;

		ote->fileobj = fileobj;
		for (i = 0; i < MAX_EVENT; i++) ote->ctx[i].fileobj = fileobj;
	}
	else {
		KdPrint(("[tdi_flt] ot_add_fileobj: Warning! reuse fileobj 0x%x\n", fileobj));

		ote->associated_fileobj = NULL;
		memset(ote->ctx, 0, sizeof(ote->ctx));
		memset(ote->local_addr, 0, sizeof(ote->local_addr));
		memset(ote->remote_addr, 0, sizeof(ote->remote_addr));
		ote->out_offset = 0;
		ote->out_oob_offset = 0;
		ote->in_offset = 0;
		ote->in_oob_offset = 0;
		ote->flt_in = 0;
		ote->flt_out = 0;
	}

	ote->signature = 'OTE ';
	ote->devobj = devobj;

	ote->type = fileobj_type;
	ote->conn_ctx = conn_ctx;
	
	status = STATUS_SUCCESS;
done:

	KeReleaseSpinLock(&g_ot_hash_guard, irql);
	return status;
}

NTSTATUS ot_del_fileobj(PFILE_OBJECT fileobj, int *fileobj_type)
{
	ULONG hash = CALC_HASH(fileobj);
	KIRQL irql;
	ot_entry_t *ote, *prev_ote;
	NTSTATUS status;

	if (!fileobj) return STATUS_INVALID_PARAMETER_1;

	KeAcquireSpinLock(&g_ot_hash_guard, &irql);

	prev_ote = NULL;
	for (ote = g_ot_hash[hash]; ote; ote = ote->next) {
		if (ote->fileobj == fileobj) break;
		prev_ote = ote;
	}

	if (!ote) {
		KdPrint(("[tdi_flt] ot_del_fileobj: fileobj 0x%x not found!\n", fileobj));
		status = STATUS_OBJECT_NAME_NOT_FOUND;
		goto done;
	}

	if (fileobj_type) *fileobj_type = ote->type;

	if (prev_ote) prev_ote->next = ote->next; else g_ot_hash[hash] = ote->next;

	datapipe_free(&ote->in_pipe);
	datapipe_free(&ote->in_oob_pipe);
	
#ifdef _DEBUG
	memset(ote, 0, sizeof(*ote)); // TEST ONLY
#endif

	free(ote);
	status = STATUS_SUCCESS;
done:

	KeReleaseSpinLock(&g_ot_hash_guard, irql);
	return status;
}

ot_entry_t *ot_find_fileobj(PFILE_OBJECT fileobj, KIRQL *irql)
{
	ULONG hash = CALC_HASH(fileobj);
	ot_entry_t *ote;
	NTSTATUS status;

	if (!fileobj) return NULL;

	if (irql) KeAcquireSpinLock(&g_ot_hash_guard, irql);

	for (ote = g_ot_hash[hash]; ote; ote = ote->next)
		if (ote->fileobj == fileobj) break;

	if (!ote) {
		KdPrint(("[tdi_flt] ot_find_fileobj: fileobj 0x%x not found!\n", fileobj));
		if (irql) KeReleaseSpinLock(&g_ot_hash_guard, *irql);
	}

	return ote;
}

//----------------------------------------------------------------------------

#define CALC_HASH_2(addrobj, conn_ctx)	CALC_HASH((ULONG)(addrobj) ^ (ULONG)(conn_ctx))

NTSTATUS ot_add_conn_ctx(PFILE_OBJECT addrobj, CONNECTION_CONTEXT conn_ctx, PFILE_OBJECT connobj)
{
	ULONG hash = CALC_HASH_2(addrobj, conn_ctx);
	KIRQL irql;
	ctx_entry_t *cte;
	NTSTATUS status;
	int i;

	KeAcquireSpinLock(&g_cte_hash_guard, &irql);
	
	for (cte = g_cte_hash[hash]; cte; cte = cte->next)
		if (cte->addrobj == addrobj && cte->conn_ctx == conn_ctx) break;

	if (!cte) {
		KdPrint(("[tdi_flt] ot_add_fileobj: Warning! reuse addrobj 0x%x, conn_ctx 0x%x\n",
			addrobj, conn_ctx));

		cte = (ctx_entry_t *)malloc(sizeof(*cte));
		if (!cte) {
			KdPrint(("[tdi_flt] ot_add_conn_ctx: malloc\n"));
			status = STATUS_NO_MEMORY;
			goto done;
		}
		cte->next = g_cte_hash[hash];
		g_cte_hash[hash] = cte;
	
		cte->addrobj = addrobj;
		cte->conn_ctx = conn_ctx;
	}

	cte->connobj = connobj;
	
	status = STATUS_SUCCESS;
done:

	KeReleaseSpinLock(&g_cte_hash_guard, irql);
	return status;
}

NTSTATUS ot_del_conn_ctx(PFILE_OBJECT addrobj, CONNECTION_CONTEXT conn_ctx)
{
	ULONG hash = CALC_HASH_2(addrobj, conn_ctx);
	KIRQL irql;
	ctx_entry_t *cte, *prev_cte;
	NTSTATUS status;

	KeAcquireSpinLock(&g_cte_hash_guard, &irql);

	prev_cte = NULL;
	for (cte = g_cte_hash[hash]; cte; cte = cte->next) {
		if (cte->addrobj == addrobj && cte->conn_ctx == conn_ctx) break;
		prev_cte = cte;
	}

	if (!cte) {
		KdPrint(("[tdi_flt] ot_del_conn_ctx: addrobj 0x%x not found!\n", addrobj));
		status = STATUS_OBJECT_NAME_NOT_FOUND;
		goto done;
	}

	if (prev_cte) prev_cte->next = cte->next; else g_cte_hash[hash] = cte->next;

	free(cte);

	status = STATUS_SUCCESS;
done:

	KeReleaseSpinLock(&g_cte_hash_guard, irql);
	return status;
}

PFILE_OBJECT ot_find_conn_ctx(PFILE_OBJECT addrobj, CONNECTION_CONTEXT conn_ctx)
{
	ULONG hash = CALC_HASH_2(addrobj, conn_ctx);
	KIRQL irql;
	ctx_entry_t *cte;
	NTSTATUS status;
	int i;
	PFILE_OBJECT result = NULL;

	KeAcquireSpinLock(&g_cte_hash_guard, &irql);
	
	for (cte = g_cte_hash[hash]; cte; cte = cte->next)
		if (cte->addrobj == addrobj && cte->conn_ctx == conn_ctx) {
			result = cte->connobj;
			break;
		}

	KeReleaseSpinLock(&g_cte_hash_guard, irql);
	return result;
}

//----------------------------------------------------------------------------

extern PDEVICE_OBJECT g_tcpfltobj, g_tcpoldobj, g_udpfltobj, g_udpoldobj;

NTSTATUS do_set_old_event_handler(ot_entry_t *ote, int event_type)
{
	NTSTATUS status;
	PIRP query_irp = NULL;
	PDEVICE_OBJECT devobj;

	if (ote->devobj == g_tcpfltobj) devobj = g_tcpoldobj;
	else if (ote->devobj == g_udpfltobj) devobj = g_udpoldobj;
	else return STATUS_UNSUCCESSFUL;

	KdPrint(("[tdi_flt] do_set_old_event_handler: devobj 0x%x, fileobj 0x%x, handler 0x%x, context 0x%x\n",
		devobj, ote->fileobj, ote->ctx[event_type].old_handler, ote->ctx[event_type].old_context));

	query_irp = TdiBuildInternalDeviceControlIrp(TDI_SET_EVENT_HANDLER, devobj,
		ote->fileobj, NULL, NULL);
	if (!query_irp) {
		KdPrint(("[tdi_flt] do_set_old_event_handler: TdiBuildInternalDeviceControlIrp\n"));
		status = STATUS_UNSUCCESSFUL;
		goto done;
	}

	TdiBuildSetEventHandler(query_irp, devobj, ote->fileobj, NULL, NULL, event_type,
		ote->ctx[event_type].old_handler, ote->ctx[event_type].old_context);

	status = IoCallDriver(devobj, query_irp);
	query_irp = NULL;

	if (status) {
		KdPrint(("[tdi_flt] do_set_old_event_handler: IoCallDriver: 0x%x\n", status));
		goto done;
	}

	// don't wait to complete

done:
	if (query_irp) IoFreeIrp(query_irp);

	return status;
}
